from typing import Optional, List

from icometrix_sdk.models.base import BackendEntity


class Role(BackendEntity):
    id: str
    name: str
    user_id: str
    entity_id: str
    entity_type: str
    second_entity_id: Optional[str] = None


class User(BackendEntity):
    id: str
    firstname: str
    lastname: str
    email: str
    hospital: str
    country: str
    language: str
    region: str
    roles: List[Role]
    blocked: bool
    verified: bool
    type: str
    note: str
    refresh_token: str
    exp: int
    token_expire_time: str
    password_setup_required: bool
    otp_active: bool



from icometrix_sdk.anonymizer.models import TagPolicy, Policy
from icometrix_sdk.anonymizer.utils import short_md5_hash, remove_if_birthday, short_sha3_hash

group_policy: Policy = {
    0x0008: TagPolicy("keep", "Group 8"),
    0x0018: TagPolicy("keep", "Acquisition: mage acquisition device and imaging procedure"),
    0x0020: TagPolicy("keep", "Relationship: Info relating the image to the location within the patient"),
    0x0028: TagPolicy("keep", "Image presentation: manner in which the image can be presented or displayed"),
    0x0029: TagPolicy("keep", "Groups specific to diffusion (b-vector and b-value, ...)"),
    0x0040: TagPolicy("remove", "Group 40"),
    0x0051: TagPolicy("keep", "DICOM HDR"),
    0x5200: TagPolicy("keep", "Multi-frame Functional Groups"),
}

policy_sha: Policy = {
    # GROUP 0x0002
    0x00020000: TagPolicy("keep", "MetaElementGroupLength"),
    0x00020001: TagPolicy("keep", "FileMetaInfoVersion"),
    0x00020002: TagPolicy("keep", "MediaStorageSOPClassUID"),
    0x00020003: TagPolicy("keep", "MediaStorageSOPInstanceUID"),
    0x00020010: TagPolicy("keep", "TransferSyntaxUID"),
    0x00020012: TagPolicy("keep", "ImplementationClassUID"),

    # GROUP 0x0008
    0x00080013: TagPolicy("round", "InstanceCreationTime"),
    0x00080014: TagPolicy("hash", "InstanceCreatorUID"),
    0x00080018: TagPolicy("hash", "SOPInstanceUID"),
    0x00080030: TagPolicy("round", "StudyTime"),
    0x00080031: TagPolicy("round", "SeriesTime"),
    0x00080032: TagPolicy("round", "AcquisitionTime"),
    0x00080033: TagPolicy("round", "ContentTime"),
    0x00080081: TagPolicy("remove", "InstitutionAddress"),
    0x00081010: TagPolicy("remove", "StationName"),
    0x00081040: TagPolicy("remove", "InstitutionalDepartmentName"),
    0x00081060: TagPolicy("remove", "NameofPhysician(s)ReadingStudy"),
    0x00081070: TagPolicy("remove", "OperatorsName"),
    0x00081155: TagPolicy("hash", "RefSOPInstanceUID"),
    0x00082111: TagPolicy("remove", "DerivationDescription"),

    # GROUP 0x0009
    0x00097772: TagPolicy("keep", "CallingAET"),

    # GROUP 0x0010
    0x00100010: TagPolicy("replace", "PatientName", replace_fn=short_sha3_hash),
    0x00100020: TagPolicy("replace", "PatientID", replace_fn=short_sha3_hash),
    0x00100021: TagPolicy("replace", "IssuerOfPatientID", replace_fn=short_sha3_hash),
    0x00100030: TagPolicy("round", "PatientBirthday"),
    0x00100032: TagPolicy("remove", "PatientBirthTime"),
    0x00100040: TagPolicy("keep", "PatientsSex"),
    0x00101000: TagPolicy("replace", "OtherPatientIDs", replace_fn=short_sha3_hash),
    0x00101010: TagPolicy("replace", "PatientAge", replace_fn=remove_if_birthday),
    0x00101030: TagPolicy("remove", "PatientWeight"),
    0x00101040: TagPolicy("remove", "PatientAddress"),

    # GROUP 0x0020
    0x00200010: TagPolicy("replace", "StudyID", replace_fn=short_sha3_hash),
    0x0020000d: TagPolicy("hash", "StudyInstanceUID"),
    0x00200052: TagPolicy("hash", "FrameOfReferenceUID"),
    0x00200200: TagPolicy("hash", "SynchronizationFrameOfReferenceUID"),
    0x0020000e: TagPolicy("hash", "SeriesInstanceUID"),
    0x00204000: TagPolicy("remove", "ImageComments"),

    # GROUP 0x0032
    0x00321060: TagPolicy("remove", "RequestedProcedureDescription"),
    0x00324000: TagPolicy("remove", "StudyComments"),

    # GROUP 0x0040
    0x0040a124: TagPolicy("hash", "UID"),

    # GROUP 0x0070
    0x0070031a: TagPolicy("hash", "FiducialUID"),

    # GROUP 0x0088
    0x00880140: TagPolicy("hash", "StorageMediaFilesetUID"),

    # GROUP 0x3006
    0x30060024: TagPolicy("hash", "ReferencedFrameOfReferenceUID"),
    0x300600c2: TagPolicy("hash", "RelatedFrameOfReferenceUID"),
    0x300a0013: TagPolicy("hash", "DoseReferenceUID"),
}

policy_md5: Policy = {
    # GROUP 0x0002
    0x00020000: TagPolicy("keep", "MetaElementGroupLength"),
    0x00020001: TagPolicy("keep", "FileMetaInfoVersion"),
    0x00020002: TagPolicy("keep", "MediaStorageSOPClassUID"),
    0x00020003: TagPolicy("keep", "MediaStorageSOPInstanceUID"),
    0x00020010: TagPolicy("keep", "TransferSyntaxUID"),
    0x00020012: TagPolicy("keep", "ImplementationClassUID"),

    # GROUP 0x0008
    0x00080013: TagPolicy("round", "InstanceCreationTime"),
    0x00080014: TagPolicy("hash", "InstanceCreatorUID"),
    0x00080018: TagPolicy("hash", "SOPInstanceUID"),
    0x00080030: TagPolicy("round", "StudyTime"),
    0x00080031: TagPolicy("round", "SeriesTime"),
    0x00080032: TagPolicy("round", "AcquisitionTime"),
    0x00080033: TagPolicy("round", "ContentTime"),
    0x00080081: TagPolicy("remove", "InstitutionAddress"),
    0x00081010: TagPolicy("remove", "StationName"),
    0x00081040: TagPolicy("remove", "InstitutionalDepartmentName"),
    0x00081060: TagPolicy("remove", "NameofPhysician(s)ReadingStudy"),
    0x00081070: TagPolicy("remove", "OperatorsName"),
    0x00081155: TagPolicy("hash", "RefSOPInstanceUID"),
    0x00082111: TagPolicy("remove", "DerivationDescription"),

    # GROUP 0x0009
    0x00097772: TagPolicy("keep", "CallingAET"),

    # GROUP 0x0010
    0x00100010: TagPolicy("replace", "PatientName", replace_fn=short_md5_hash),
    0x00100020: TagPolicy("replace", "PatientID", replace_fn=short_md5_hash),
    0x00100021: TagPolicy("replace", "IssuerOfPatientID", replace_fn=short_md5_hash),
    0x00100030: TagPolicy("round", "PatientBirthday"),
    0x00100032: TagPolicy("remove", "PatientBirthTime"),
    0x00100040: TagPolicy("keep", "PatientsSex"),
    0x00101000: TagPolicy("replace", "OtherPatientIDs", replace_fn=short_md5_hash),
    0x00101010: TagPolicy("replace", "PatientAge", replace_fn=remove_if_birthday),
    0x00101030: TagPolicy("remove", "PatientWeight"),
    0x00101040: TagPolicy("remove", "PatientAddress"),

    # GROUP 0x0020
    0x00200010: TagPolicy("replace", "StudyID", replace_fn=short_md5_hash),
    0x0020000d: TagPolicy("hash", "StudyInstanceUID"),
    0x00200052: TagPolicy("hash", "FrameOfReferenceUID"),
    0x00200200: TagPolicy("hash", "SynchronizationFrameOfReferenceUID"),
    0x0020000e: TagPolicy("hash", "SeriesInstanceUID"),
    0x00204000: TagPolicy("remove", "ImageComments"),

    # GROUP 0x0032
    0x00321060: TagPolicy("remove", "RequestedProcedureDescription"),
    0x00324000: TagPolicy("remove", "StudyComments"),

    # GROUP 0x0040
    0x0040a124: TagPolicy("hash", "UID"),

    # GROUP 0x0070
    0x0070031a: TagPolicy("hash", "FiducialUID"),

    # GROUP 0x0088
    0x00880140: TagPolicy("hash", "StorageMediaFilesetUID"),

    # GROUP 0x3006
    0x30060024: TagPolicy("hash", "ReferencedFrameOfReferenceUID"),
    0x300600c2: TagPolicy("hash", "RelatedFrameOfReferenceUID"),
    0x300a0013: TagPolicy("hash", "DoseReferenceUID"),
}



from enum import Enum


class Region(Enum):
    EU = "https://icobrain-eu.icometrix.com"
    US = "https://icobrain-us.icometrix.com"
    CA = "https://icobrain-ca.icometrix.com"
    AU = "https://icobrain-au.icometrix.com"
    STUDIES = "https://studies.icometrix.com"



"""
An example on how to authenticate using an email address and password
"""

import os

from icometrix_sdk import IcometrixApi, PasswordAuthentication
from icometrix_sdk.utils.requests_api_client import RequestsApiClient

if __name__ == '__main__':
    SERVER = "https://icobrain-test.icometrix.com"

    auth = PasswordAuthentication("example@company.com", os.environ["PASSWORD"])
    client = RequestsApiClient(SERVER, auth)

    ico_api = IcometrixApi(client)

    me = ico_api.profile.who_am_i()

    print(me.email)



import logging

from pydicom import Dataset, DataElement

from icometrix_sdk.anonymizer.exceptions import PolicyException
from icometrix_sdk.anonymizer.hash_factory import HashMethod
from icometrix_sdk.anonymizer.models import Policy, TagPolicy
from icometrix_sdk.anonymizer.utils import remove_tag, replace_tag, hash_tag, _is_pixel_data, round_tag, \
    add_de_identification_tags, is_tag, is_group, empty_tag
from icometrix_sdk.logger import logger_name

logger = logging.getLogger(logger_name)


class Anonymizer:
    default_policy: TagPolicy = TagPolicy("empty", "Default")

    def __init__(self, policy: Policy, group_policy: Policy, hash_algo: HashMethod):
        for tag in policy:
            if not is_tag(tag):
                raise PolicyException("Tag policy contains an invalid tag")

        for group in group_policy:
            if not is_group(group):
                raise PolicyException("Group policy contains an invalid group")

        self.policy = policy
        self.group_policy = group_policy
        self.hash_algo = hash_algo

    def anonymize(self, dataset: Dataset) -> Dataset:
        for element in dataset.iterall():
            # Keep the pixel data
            if _is_pixel_data(element.tag):
                continue

            # Apply the tag policy
            elif element.tag in self.policy:
                tag_policy = self.policy[element.tag]
                self._apply_policy_to_tag(element, tag_policy, dataset)

            # Apply the group policy
            elif element.tag.group in self.group_policy:
                tag_policy = self.group_policy[element.tag.group]
                self._apply_policy_to_group(element, tag_policy, dataset)

            # Apply the default policy
            else:
                self._apply_policy_to_tag(element, self.default_policy, dataset)

        return add_de_identification_tags(dataset)

    def _apply_policy_to_group(self, element: DataElement, tag_policy: TagPolicy, dataset: Dataset):
        try:
            self._apply_policy_to_tag(element, tag_policy, dataset)
        except (AttributeError, ValueError):
            logger.debug("Failed to apply group action '%d' to %s %s.", tag_policy.action,
                         element.tag, element.name)
            return

    def _apply_policy_to_tag(self, element: DataElement, tag_policy: TagPolicy, dataset: Dataset):
        logger.debug('%d %s: %s', element.tag, element.name, tag_policy.action)

        if tag_policy.action == "keep":
            return
        elif tag_policy.action == "empty":
            empty_tag(element)
        elif tag_policy.action == "remove":
            remove_tag(element, dataset)
        elif tag_policy.action == "replace":
            replace_tag(element, dataset, tag_policy.replace_fn)
        elif tag_policy.action == "hash":
            hash_tag(element, self.hash_algo)
        elif tag_policy.action == "round":
            round_tag(element)
        else:
            raise ValueError("Unknown tag policy action")



import logging

from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.series_entity import SeriesEntity
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class Series:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_all_for_study(self, study_uri: str, **kwargs) -> PaginatedResponse[SeriesEntity]:
        """
        List all series for a study

        :param study_uri: the uri of the study
        :return: A Paginated response containing series
        """
        page = self._api.get(f"{study_uri}/series", **kwargs)
        return PaginatedResponse[SeriesEntity](**page)



class IcometrixException(Exception):
    pass


class IcometrixHttpException(IcometrixException):
    fmt = "An HTTP Client raised an unhandled exception: {error}"


class IcometrixParseException(IcometrixException):
    fmt = "Failed to parse HTTP response body"


class IcometrixConfigException(IcometrixException):
    fmt = "Invalid configuration"


class IcometrixAuthException(IcometrixException):
    fmt = "Authentication failed"


class IcometrixDataImportException(IcometrixException):
    fmt = "Dicom import failed"


class IcometrixInvalidInputDataException(IcometrixException):
    fmt = "Data validation failed"


class WaitTimeoutException(IcometrixException):
    ...


class IcometrixUnusableReportException(IcometrixException):
    pass



from __future__ import annotations

import os
import sys
from datetime import date

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

root_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
sys.path.insert(0, root_path)

from icometrix_sdk._version import __version__

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "icometrix-sdk"
copyright = f"{date.today().year}, icometrix"
author = "icometrix"
release = __version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autodoc",
    "sphinxcontrib.mermaid"
]

# sphinx.ext.autodoc options
autodoc_default_options = {
    'exclude-members': 'model_config, model_fields, model_post_init'  # exclude pydantic internal fields
}

templates_path = ["_templates"]
exclude_patterns = []

# Show typehints as content of the function or method
autodoc_typehints = "description"

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
# html_static_path = ["_static"]



from typing import Optional, Union, List, Literal

from pydantic import BaseModel

from icometrix_sdk.models.base import BackendEntity

CustomerReportStatus = Literal["Imported", "Scheduled", "Processing", "QualityControl", "Finished"]

class CustomerReportFile(BaseModel):
    uri: str
    name: str
    type: str

    def __str__(self):
        return (
            f"Name: {self.name} | "
            f"Type: {self.type}"
        )

    def __repr__(self):
        return str(self)


class CustomerReportEntity(BackendEntity):
    cost: Optional[Union[str, float]] = None
    status: str
    study_id: str
    upload_id: Optional[str] = None
    patient_id: str
    project_id: str
    study_date: Optional[str] = None
    report_type: Optional[str] = None
    patient_name: Optional[str] = None
    project_name: str
    report_files: List[CustomerReportFile]
    report_status: Optional[str] = None
    project_country: str
    accession_number: Optional[str] = None
    dicom_patient_id: Optional[str] = None
    customer_result_id: Optional[str] = None
    pipeline_result_id: Optional[str] = None
    study_instance_uid: str
    icobrain_report_type: str
    report_translated_remarks: Optional[str] = None

    def __str__(self):
        return (
            f"Report ID: {self.id[:8]} | "
            f"Status: {self.status} | "
            f"Report Status: {self.report_status} | "
            f"Report Type: {self.icobrain_report_type}"
        )

    def __repr__(self):
        return str(self)



import os

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.utils.paginator import get_paginator

PROJECT_ID = "uuid"
DICOM_DIR_PATH = "<path>"

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    for reports in get_paginator(ico_api.customer_reports.get_all, page_size=30, project_id=PROJECT_ID):
        for report in reports:
            if report.status == "Finished":
                print(report.study_instance_uid, report.report_status)



from pydicom import Dataset, DataElement

from icometrix_sdk.anonymizer.config import PATIENT_IDENTITY_REMOVED_TAG, DE_IDENTIFICATION_METHOD_TAG, \
    PRIVATE_ICOMETRIX_GROUPS
from icometrix_sdk.anonymizer.utils import _is_pixel_data


def _ignore_tag(tag: int) -> bool:
    # Tags that can be ignored during testing

    return ((_is_pixel_data(tag)
             or tag in (PATIENT_IDENTITY_REMOVED_TAG, DE_IDENTIFICATION_METHOD_TAG))
            or (tag >> 16 in PRIVATE_ICOMETRIX_GROUPS))


def _split_group_tags(dataset: Dataset, group: int) -> tuple[list[DataElement], list[DataElement]]:
    # A helper function to separate elements in a group, and not in a group (often needed for testing)
    in_group: list[DataElement] = []
    outside_group: list[DataElement] = []
    for element in dataset:
        tag = element.tag
        if tag.group == group:
            in_group.append(element)
        else:
            outside_group.append(element)
    return in_group, outside_group


def replace_by_value(value: str | int):
    def replace_vl(el: DataElement, _):
        el.value = value

    return replace_vl



from dataclasses import dataclass
from typing import Literal, Callable

from pydicom import DataElement, Dataset

Action = Literal["keep", "empty", "remove", "replace", "hash", "round"]
HashAlgo = Literal["sha3", "md5", "short_md5"]
ReplaceFn = Callable[[DataElement, Dataset], None]


@dataclass
class TagPolicy:
    """
     represents a DICOM tag anonymization policy, defining how individual tags or groups should be anonymized.

    :param: Action to be performed on the tag (keep, remove, replace, hash, round)
    :param: Description of the tag anonymization policy.
    :param: Replace Function to be used when the action is "replace" (optional).

    Actions:
    - keep: Keep the original tag value.
    - remove: Remove the tag from the DICOM dataset.
    - replace: Replace the tag value with a specified :attr:`~icometrix_sdk.anonymizer.models.TagPolicy.value`.
    - hash: Hash the tag value using a specified algorithm.
    - round: Round the tag value to the nearest value. (currently only dates VRs are supported)

    """
    action: Action
    description: str

    # value: int | str = None
    replace_fn: ReplaceFn = None


Policy = dict[int, TagPolicy]



from datetime import datetime, timezone
from typing import Generic, TypeVar, List, Optional, Iterator, Literal

from pydantic import BaseModel, BeforeValidator
from typing_extensions import Annotated

T = TypeVar("T")

DicomModality = Literal[
    "CR", "CT", "MR", "US", "OT", "BI", "CD", "DD", "DG", "ES", "LS", "PT", "RG", "ST", "TG", "XA",
    "RF", "HC", "DX", "NM", "MG", "IO", "PX", "GM", "SM", "XC", "PR", "AU", "EPS", "HD", "SR", "OP",
    "IVUS", "SMR", "RTIMAGE", "RTDOSE", "RTSTRUCT", "RTPLAN", "RTRECORD"]


def utc_datetime_parser(v: str | datetime | None) -> datetime | None:
    if not v:
        return None
    if isinstance(v, datetime):
        return v.replace(tzinfo=timezone.utc)

    datetime_format = "%Y-%m-%dT%H:%M:%S.%f"
    if "+" in v:
        datetime_format = "%Y-%m-%dT%H:%M:%S.%f%z"

    input_datetime = datetime.strptime(v, datetime_format)
    return input_datetime.replace(tzinfo=timezone.utc)


# Does not work without the None
utc_datetime = Annotated[
    datetime | None,
    BeforeValidator(utc_datetime_parser),
]


class BackendEntity(BaseModel):
    id: str
    uri: Optional[str] = None
    update_timestamp: utc_datetime
    creation_timestamp: utc_datetime


class PaginatedResultSet(BaseModel):
    count: int
    offset: int
    limit: int


class PaginatedMetaData(BaseModel):
    result_set: PaginatedResultSet


class PaginatedResponse(BaseModel, Generic[T]):
    meta_data: PaginatedMetaData
    results: List[T]
    _itr_index: int

    def has_next(self) -> bool:
        """
        Returns True if there is a next page in the backend

        :return: bool
        """
        result_set = self.meta_data.result_set
        return int(result_set.count) > (int(result_set.offset) + int(result_set.limit))

    def __iter__(self) -> Iterator[T]:
        return iter(self.results)

    def __getitem__(self, index) -> T:
        return self.results[index]

    def __next__(self) -> T:
        if len(self.results) < self._itr_index:
            raise StopIteration

        item: T = self.results[self._itr_index]
        self._itr_index += 1
        return item

    def __len__(self) -> int:
        return len(self.results)



from typing import Optional, List, Union, Literal

from pydantic import BaseModel

from icometrix_sdk.models.base import BackendEntity

SuggestedReportTypes = Literal["icobrain_ms", "icobrain_dm", "icobrain_tbi"]


class StartUploadDto(BaseModel):
    icobrain_report_type: Union[SuggestedReportTypes, str]
    force_reprocessing: bool = False


class UploadEntity(BackendEntity):
    status: Optional[str] = None
    folder_uri: str
    type: str
    compressed: Optional[Union[str, bool]] = None
    logs: List[str]
    errors: List[str]
    icobrain_report_type: str
    retry_count: Optional[Union[str, int]] = None
    project_id: str

    def __str__(self):
        latest_log = self.logs[-1] if self.logs else "-"
        errors = ", ".join(self.errors) if self.errors else "-"
        log_message = (
            f"Upload ID: {self.id[:8]} | "
            f"Status: {self.status} | "
            f"Logs: {latest_log} | "
            f"Errors: {errors}"
        )
        return log_message

    def __repr__(self):
        return str(self)


class StudyUploadEntity(BackendEntity):
    project_id: str
    patient_id: str
    study_id: str
    upload_id: str


class UploadEntityFiles(BaseModel):
    files: List[str]



from icometrix_sdk.models.base import BackendEntity, utc_datetime


class SeriesEntity(BackendEntity):
    note: str
    modality: str
    study_id: str
    echo_time: str  # available after preprocessing
    slice_gap: float  # available after preprocessing
    flip_angle: str
    image_size: list[int]  # available after preprocessing
    patient_id: str
    project_id: str
    quarantine: str
    voxel_size: list[float]  # available after preprocessing
    series_date: str
    series_time: str
    series_type: str  # available after preprocessing
    pixel_spacing: str
    sequence_name: str
    series_number: str
    inversion_time: str
    images_in_nifti: int  # available after preprocessing
    repetition_time: str
    slice_thickness: float  # available after preprocessing
    sequence_variant: str
    receive_coil_name: str
    scanning_sequence: str
    slice_orientation: str
    body_part_examined: str
    imported_timestamp: utc_datetime
    series_description: str
    transmit_coil_name: str
    series_instance_uid: str
    images_in_acquisition: str | int
    magnetic_field_strength: str
    maximum_slice_increment: float  # available after preprocessing
    expected_images_in_acquisition: str



"""A library that provides a Python interface to the Icometrix API"""
import os
from typing import Optional

from icometrix_sdk.authentication import PasswordAuthentication, AuthenticationMethod, get_auth_method
from icometrix_sdk.exceptions import IcometrixConfigException
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.upload_entity import StartUploadDto
from icometrix_sdk.resources.customer_reports import CustomerReports
from icometrix_sdk.resources.customer_results import CustomerResults
from icometrix_sdk.resources.patients import Patients
from icometrix_sdk.resources.pipeline_results import PipelineResults
from icometrix_sdk.resources.profile import Profile
from icometrix_sdk.resources.projects import Projects
from icometrix_sdk.resources.series import Series
from icometrix_sdk.resources.studies import Studies
from icometrix_sdk.resources.uploads import Uploads
from icometrix_sdk.utils.api_client import ApiClient
from icometrix_sdk.utils.regions import Region
from icometrix_sdk.utils.requests_api_client import RequestsApiClient


def get_api_client() -> RequestsApiClient:
    """
    Create an RequestsApiClient with the 'API_HOST' environment variable as server
    """

    region = os.getenv("REGION")
    if region in Region:
        api_host = Region[region].value
    else:
        api_host = os.getenv("API_HOST")

    if not api_host:
        raise IcometrixConfigException("Please set the 'API_HOST' environment variable")

    auth_method = get_auth_method()
    return RequestsApiClient(api_host, auth_method)


class IcometrixApi:
    """
    A Python interface for the icometrix API.

    :param api_client:
        An instance of an ApiClient
    """

    profile: Profile
    projects: Projects
    patients: Patients
    studies: Studies
    series: Series
    uploads: Uploads
    customer_results: CustomerResults
    customer_reports: CustomerReports

    _api_client: ApiClient

    def __init__(self, api_client: Optional[ApiClient] = None):
        self._api_client = api_client
        if not self._api_client:
            self._api_client = get_api_client()

        self.profile = Profile(self._api_client)
        self.projects = Projects(self._api_client)
        self.patients = Patients(self._api_client)
        self.studies = Studies(self._api_client)
        self.series = Series(self._api_client)
        self.uploads = Uploads(self._api_client)
        self.pipeline_results = PipelineResults(self._api_client)
        self.customer_results = CustomerResults(self._api_client)
        self.customer_reports = CustomerReports(self._api_client)

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass
        # todo force logout
        # if self._api_client.auth:
        #     self._api_client.auth.disconnect(self._api_client)

    #         """
    #         Download a file created by icobrain for a customer_report
    #
    #         :param customer_report_file: The file entity you want to download
    #         :param out_path: A path to write the file to
    #         :return:
    #         """

    def process_dicom_directory(self, project_id: str, input_dir: str, out_dir: str, params: StartUploadDto):
        """
        The process_dicom_directory function provides a high-level interface for handling the complete workflow of
        processing DICOM files and downloading the results.

        :param project_id (str): The UUID of the project to which the DICOM files belong.
        :param input_dir (str): The path to the directory containing the DICOM files to be processed.
        :param out_dir (str): The path to the directory where output data will be stored.
        :param params (StartUploadDto): An instance of StartUploadDto containing parameters for the upload.
        :return:
        """
        # Get the project, to make sure its there (will throw a 404 in case the project is not found)
        project = self.projects.get_one_by_id(project_id)

        # Upload a directory of DICOMs
        upload = self.uploads.upload_dicom_dir(project.id, input_dir, params)

        # Wait for data to be imported
        upload = self.uploads.wait_for_data_import(upload.folder_uri)

        # e.g. No DICOMs found in the upload
        if upload.errors:
            raise Exception(f"Upload failed: {', '.join(upload.errors)}")

        # Get imported studies
        studies_in_upload = self.uploads.get_studies_for_upload(upload_folder_uri=upload.folder_uri)
        if not studies_in_upload:
            raise Exception("No valid studies uploaded.")

        for study_in_upload in studies_in_upload:

            # Find the study
            study = self.studies.get_one(study_in_upload.project_id,
                                         study_in_upload.patient_id,
                                         study_in_upload.study_id)

            # Get reports for this study
            customer_reports = self.customer_reports.get_all_for_study(study_uri=study.uri)

            # Wait for the reports to finish and download the results
            finished_customer_reports = self.customer_reports.wait_for_results(list(customer_reports))
            for customer_report in finished_customer_reports:
                out_path = os.path.join(out_dir, customer_report.id)
                self.customer_reports.download_customer_report_files(customer_report, out_path)



import logging

from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.pipeline_result_entity import PipelineResultEntity
from icometrix_sdk.utils.paginator import get_paginator
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class PipelineResults:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_all_for_study(self, study_uri: str, **kwargs) -> PaginatedResponse[PipelineResultEntity]:
        """
        Get al pipeline reports for a project

        :param study_uri: The uri of a study
        :return: A Paginated response containing pipeline-results
        """

        study_uri = study_uri.replace("/v1/", "/v2/")
        page = self._api.get(f"{study_uri}/pipeline-results", **kwargs)
        return PaginatedResponse[PipelineResultEntity](**page)

    def get_one_for_job(self, study_uri: str, job_id: str) -> PipelineResultEntity | None:
        """
        Get a pipeline result for a job

        :param study_uri: The uri of a study
        :param job_id: The id of the job
        """
        page = get_paginator(self.get_all_for_study, study_uri=study_uri)
        for pipeline_results in page:
            for pipeline_result in pipeline_results:
                if pipeline_result.job_id == job_id:
                    return pipeline_result

    def get_one(self, pipeline_result_uri: str) -> PipelineResultEntity:
        """
        Get a single customer-result based on the customer-result uri

        :param pipeline_result_uri: the uri of the pipeline-result
        :return: A single pipeline-result or 404
        """
        resp = self._api.get(pipeline_result_uri)
        return PipelineResultEntity(**resp)



"""
An example on how to implement custom authentication logic
"""

from icometrix_sdk import IcometrixApi, AuthenticationMethod
from icometrix_sdk.utils.requests_api_client import RequestsApiClient

SERVER = "https://icobrain-test.icometrix.com"


class FileAuth(AuthenticationMethod):
    def __init__(self, secret_password_file_path: str):
        self.secret_password_file_path = secret_password_file_path

    def connect(self, api: RequestsApiClient):
        with open(self.secret_password_file_path) as secret:
            email = secret.readlines()[0],
            password = secret.readlines()[1]
            data = {"email": email, "password": password}
        api.post("/authentication-service/api/v1/sessions", data=data)

    def disconnect(self, api: RequestsApiClient):
        pass


if __name__ == '__main__':
    auth = FileAuth("/mount/secret.txt")
    client = RequestsApiClient(SERVER, auth)
    ico_api = IcometrixApi(client)



import logging

from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.study_entity import StudyEntity
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class Studies:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_one(self, project_id: str, patient_id: str, study_id: str) -> StudyEntity:
        """
        Get a single upload entry based on the upload uri

        :param project_id: the ID of the project
        :param patient_id: the ID of the patient
        :param study_id: the ID of the study
        :return: A single study or 404
        """
        resp = self._api.get(f"/storage-service/api/v1/projects/{project_id}/patients/{patient_id}/studies/{study_id}")
        return StudyEntity(**resp)

    def get_all_for_patient(self, patient_uri: str, **kwargs) -> PaginatedResponse[StudyEntity]:
        """
        List all studies for a patient

        :param: The uri of the patient
        :return: A Paginated response containing studies
        """
        page = self._api.get(f"{patient_uri}/patients", **kwargs)
        return PaginatedResponse[StudyEntity](**page)



import logging
import os
from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.models.upload_entity import StartUploadDto

PROJECT_ID = "uuid"
REPORT_TYPE = "icobrain_ms"

# A directory containing a T1 and FLAIR scan, as required for icobrain_ms for more info see HCP manual
DICOM_DIR_PATH = "<dir_path>"

logging.basicConfig(level=logging.INFO)

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    # Get the project, to make sure its there (will throw a 404 in case the project is not found)
    project = ico_api.projects.get_one_by_id(PROJECT_ID)

    # Upload a directory of DICOMs
    data = StartUploadDto(icobrain_report_type=REPORT_TYPE)
    upload = ico_api.uploads.upload_dicom_dir(PROJECT_ID, DICOM_DIR_PATH, data)

    # Wait for data to be imported
    upload = ico_api.uploads.wait_for_data_import(upload.folder_uri)

    # e.g. No DICOMs found in the upload
    if upload.errors:
        raise Exception(f"Upload failed: {', '.join(upload.errors)}")

    # Get imported studies
    studies_in_upload = ico_api.uploads.get_studies_for_upload(upload_folder_uri=upload.folder_uri)
    if not studies_in_upload:
        raise Exception("No valid studies uploaded.")

    # For the demo we just use a single study
    study_in_upload = studies_in_upload[0]
    study = ico_api.studies.get_one(study_in_upload.project_id, study_in_upload.patient_id, study_in_upload.study_id)

    # Get reports for this study
    customer_reports = ico_api.customer_reports.get_all_for_study(study_uri=study.uri)

    # Wait for the reports to finish and download the results
    finished_customer_reports = ico_api.customer_reports.wait_for_results(list(customer_reports))
    for customer_report in finished_customer_reports:
        ico_api.customer_reports.download_customer_report_files(customer_report, f"./{customer_report.id}")



import logging

from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.customer_result_entity import CustomerResultEntity
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class CustomerResults:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_all_for_study(self, study_uri: str, **kwargs) -> PaginatedResponse[CustomerResultEntity]:
        """
        Get al customer results for a study

        :param study_uri: The uri of a study
        :return: A Paginated response containing customer-results
        """

        page = self._api.get(f"{study_uri}/customer-results", **kwargs)
        return PaginatedResponse[CustomerResultEntity](**page)

    def get_all_for_patient(self, patient_uri: str, **kwargs) -> PaginatedResponse[CustomerResultEntity]:
        """
        Get al customer results for a patient

        :param patient_uri: The uri of a patient
        :return: A Paginated response containing customer-results
        """

        page = self._api.get(f"{patient_uri}/customer-results", **kwargs)
        return PaginatedResponse[CustomerResultEntity](**page)

    def get_all_for_pipeline_result(self, pipeline_result_uri: str, **kwargs) -> \
            PaginatedResponse[CustomerResultEntity]:
        """
        Get al customer reports for a pipeline-result

        :param pipeline_result_uri: The uri of a pipeline-result
        :return: A Paginated response containing customer-results
        """

        page = self._api.get(f"{pipeline_result_uri}/customer-results", **kwargs)
        return PaginatedResponse[CustomerResultEntity](**page)

    def get_one(self, customer_result_uri: str) -> CustomerResultEntity:
        """
        Get a single customer-result based on the customer-result uri

        :param customer_result_uri: the uri of the customer-result
        :return: A single customer-result or 404
        """
        resp = self._api.get(customer_result_uri)
        return CustomerResultEntity(**resp)



import re
from typing import TypedDict


class StudyUriUuids(TypedDict):
    project_id: str
    patient_id: str
    study_id: str


def extract_uuids(uri: str) -> list[str]:
    # Regular expression to match UUIDs
    uuid_pattern = re.compile(r"[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}")
    # Find all matches in the URI
    return uuid_pattern.findall(uri)


def extract_uuids_from_study_uri(study_uri: str) -> StudyUriUuids:
    uuids = extract_uuids(study_uri)
    if len(uuids) != 3:
        raise Exception("invalid study_uri")

    return {"project_id": uuids[0], "patient_id": uuids[1], "study_id": uuids[2]}



from datetime import timedelta
from typing import Literal, Optional

from pydantic import BaseModel

from icometrix_sdk.models.base import BackendEntity, utc_datetime
from icometrix_sdk.utils.time_diff import time_difference


class JobParameters(BaseModel):
    pipeline_json_file: str
    pipeline_parameters_json_file: str
    run_parameters_json_string: str
    docker_image: str


class JobEntity(BackendEntity):
    status: str
    parameters: JobParameters
    comment: str
    # console_log: str
    job_type: str  # Optional[Literal['cross_processsing', 'long_processsing', 'preprocessing']] = None
    finished_timestamp: Optional[utc_datetime] = None
    reschedule_counter: Optional[str | int] = None
    job_mode: Optional[str] = None
    study_ids: list[str]
    project_id: str
    patient_id: str
    operator_id: str
    reference_job_id: Optional[str] = None

    def running_time(self) -> timedelta:
        if not self.finished_timestamp:
            return time_difference(self.creation_timestamp)
        return self.finished_timestamp - self.creation_timestamp

    def __str__(self):
        return (
            f"Job ID: {self.id[:8]} | "
            f"Status: {self.status} | "
            f"Job Type: {self.job_type} | "
            f"Running Time: {self.running_time()}"
        )

    def __repr__(self):
        return str(self)


class StartPreProcessing(BaseModel):
    # patient_id: str
    # project_id: str
    # study_id: str
    # study_uri: str
    # docker_image: str
    icobrain_report_type: Literal['icobrain_ms', 'icobrain_dm']
    icobrain_report_language: Optional[str] = None
    # report_languages: list[str]
    force_reprocessing: Optional[bool] = False
    disable_quality_control: Optional[bool] = False
    # type: str


class StartCrossProcessing(BaseModel):
    # patient_id: str
    # project_id: str
    # study_id: str
    # study_uri: str
    t1_uri: Optional[str] = None
    t1_post_uri: Optional[str] = None
    flair_uri: Optional[str] = None
    ct_uri: Optional[str] = None
    t2star_uri: Optional[str] = None
    t2_uri: Optional[str] = None
    pd_uri: Optional[str] = None
    swi_uri: Optional[str] = None
    mt_on_uri: Optional[str] = None
    mt_off_uri: Optional[str] = None
    dmri_uri: Optional[str] = None
    t1_contrast_enhanced: Optional[str] = None
    # comment: Optional[str]
    # docker_image: str
    # resource_requirements: Resourcerequirements;
    icobrain_report_type: Literal['icobrain_ms', 'icobrain_dm']
    icobrain_report_language: Optional[str] = None
    # report_languages: list[str]
    force_reprocessing: Optional[bool] = False
    disable_quality_control: Optional[bool] = False
    # type: str
    # quality_control_options: Resourcerequirements;
    # scientific_options: Resourcerequirements;
    # job_mode: str



import logging
import os
from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.models.upload_entity import StartUploadDto

PROJECT_ID = "<project_uuid>"
REPORT_TYPE = "<report_type>"

# A directory containing a T1 and FLAIR scan, as required for icobrain_ms for more info see HCP manual
DICOM_DIR_PATH = "<input_dir_path>"
OUTPUT_DIR = "<output_dir_path>"

logging.basicConfig(level=logging.INFO)

if __name__ == '__main__':
    os.environ["API_HOST"] = ""

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    processing_data = StartUploadDto(icobrain_report_type=REPORT_TYPE)
    ico_api.process_dicom_directory(PROJECT_ID, DICOM_DIR_PATH, OUTPUT_DIR, processing_data)


import os

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.models.upload_entity import StartUploadDto

PROJECT_ID = "uuid"
DICOM_DIR_PATH = "<dir_path>"

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    # Get the project, to make sure its there (will throw a 404 in case the project is not found)
    project = ico_api.projects.get_one_by_id(PROJECT_ID)

    # Start an upload
    data = StartUploadDto(icobrain_report_type="icobrain_ms")
    upload = ico_api.uploads.start_upload(PROJECT_ID, data)

    # Will walk through all files/subdirectories and upload all files
    ico_api.uploads.upload_all_files_in_dir(upload.uri, DICOM_DIR_PATH)

    # Once all files have been uploaded, signal that they are all there and start the import/processing
    upload = ico_api.uploads.complete_upload(upload.uri)



from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.project_entity import ProjectEntity
from icometrix_sdk.utils.requests_api_client import ApiClient


class Projects:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_all(self, **kwargs) -> PaginatedResponse[ProjectEntity]:
        """
        List projects that the current user has access to

        :return: A Paginated response containing projects
        """
        no_settings = {"no-settings": "true"}
        if "params" in kwargs:
            kwargs["params"].update(no_settings)
        else:
            kwargs["params"] = no_settings
        page = self._api.get("/storage-service/api/v2/projects", **kwargs)
        return PaginatedResponse[ProjectEntity](**page)

    def get_one(self, project_uri: str) -> ProjectEntity:
        """
        Get a single project based on the project uri

        :param project_uri: the uri of the project
        :return: A single project or 404
        """
        params = {"no-settings": "true"}
        resp = self._api.get(project_uri, params=params)
        return ProjectEntity(**resp)

    def get_one_by_id(self, project_id: str) -> ProjectEntity:
        """
        Get a single project based on the project ID

        :param project_id: the ID of the project
        :return: A single project or 404
        """
        params = {"no-settings": "true"}
        resp = self._api.get(f"/storage-service/api/v1/projects/{project_id}", params=params)
        return ProjectEntity(**resp)



from typing import Optional, Union

from icometrix_sdk.models.base import BackendEntity, utc_datetime


class PatientEntity(BackendEntity):
    icon: Optional[str] = None
    alias: Optional[str] = None
    label: Optional[str] = None
    cohort: Optional[str] = None
    country: Optional[str] = None
    hospital: Optional[str] = None
    lastname: Optional[str] = None
    firstname: Optional[str] = None
    birth_year: Optional[Union[str, int]] = None
    patient_id: str
    project_id: str
    invite_code: Optional[str] = None
    patient_sex: Optional[str] = None
    patient_name: str
    project_name: Optional[str] = None
    trial_excluded: Optional[Union[str, bool]] = None
    project_country: Optional[str] = None
    msbase_patient_id: Optional[str] = None
    unique_patient_id: Optional[str] = None
    imported_timestamp: Optional[utc_datetime] = None
    patient_birth_date: str
    patient_handedness: Optional[str] = None
    encrypted_patient_id: Optional[str] = None
    encrypted_patient_name: Optional[str] = None
    encrypted_patient_birth_date: Optional[str] = None



import logging
import os

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.models.upload_entity import StartUploadDto

PROJECT_ID = "pid"

# A directory containing a T1 and FLAIR scan, as required for icobrain_ms for more info see HCP manual
DICOM_DIR_PATH = "..."

logging.basicConfig(level=logging.INFO)

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    # Get the project, to make sure its there (will throw a 404 in case the project is not found)
    project = ico_api.projects.get_one_by_id(PROJECT_ID)

    # Upload a directory of DICOMs
    data = StartUploadDto(icobrain_report_type="icobrain_ms")
    upload = ico_api.uploads.upload_dicom_dir(PROJECT_ID, DICOM_DIR_PATH, data)

    # Wait for data to be imported
    upload = ico_api.uploads.wait_for_data_import(upload.folder_uri)

    # Get imported studies
    studies_in_upload = ico_api.uploads.get_studies_for_upload(upload_folder_uri=upload.folder_uri)

    for uploaded_study in studies_in_upload:
        study = ico_api.studies.get_one(uploaded_study.project_id, uploaded_study.patient_id, uploaded_study.study_id)

        # Add some validation logic for each study e.g. find series
        series = ico_api.series.get_all_for_study(study.uri)
        # ...



import pytest
from pydicom import DataElement, Dataset

from icometrix_sdk.anonymizer.tests.utils import replace_by_value
from icometrix_sdk.anonymizer.utils import replace_tag


def test_replace_existing_value():
    expected = "John^Doe"
    elem1 = DataElement(0x00100010, "PN", "Jane^Doe")
    replace_tag(elem1, Dataset(), replace_by_value(expected))

    assert elem1.value == expected


def test_replace_empty_value():
    expected = "072603.066406"
    elem1 = DataElement(0x00080050, "SH", "4753014.1")
    replace_tag(elem1, Dataset(), replace_by_value(expected))

    assert elem1.value == expected


def test_replace_with_to_long_value():
    expected = "072603.066406.to.long.value"
    elem1 = DataElement(0x00080050, "SH", "4753014.1")
    with pytest.raises(ValueError):
        replace_tag(elem1, Dataset(), replace_by_value(expected))


def test_replace_with_invalid_type_value():
    expected = 19
    elem1 = DataElement(0x00080020, "DA", "19700101")
    with pytest.raises(ValueError):
        replace_tag(elem1, Dataset(), replace_by_value(expected))



from urllib3 import encode_multipart_formdata


def create_multipart(fields, boundary=None):
    """
    Wrapper around urllib3 "encode_multipart_formdata"

    :param fields:
        Dictionary of fields or list of (key, :class:`~urllib3.fields.RequestField`).

    :param boundary:
        If not specified, then a random boundary will be generated using
        :func:`urllib3.filepost.choose_boundary`.
    """
    data, content_type = encode_multipart_formdata(fields, boundary)
    headers = {"Content-Type": content_type, 'Content-Length': str(len(data))}
    return data, headers



from abc import abstractmethod, ABC


class ApiClient(ABC):
    """
    Interface for API clients, used to make HTTP calls to the icometrix API
    """

    @abstractmethod
    def get(self, uri: str, **kwargs) -> dict:
        pass

    @abstractmethod
    def post(self, uri: str, data: dict, **kwargs) -> dict:
        pass

    @abstractmethod
    def put(self, uri: str, data: dict, **kwargs) -> dict:
        pass

    @abstractmethod
    def delete(self, uri: str, **kwargs):
        pass

    @abstractmethod
    def put_file(self, uri: str, fields, **kwargs):
        pass

    @abstractmethod
    def stream_file(self, uri: str, out_path: str, **kwargs):
        pass



import logging
import os
from time import sleep
from typing import Optional, Dict, List

from icometrix_sdk.exceptions import IcometrixDataImportException
from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.customer_report_entity import CustomerReportEntity, CustomerReportFile
from icometrix_sdk.utils.paginator import get_paginator
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class CustomerReports:
    def __init__(self, api: ApiClient, polling_interval=2):
        self._api = api
        self.polling_interval = polling_interval

    def get_all(self, project_id: str, **kwargs) -> PaginatedResponse[CustomerReportEntity]:
        """
        Get al customer reports for a project

        :param project_id: The ID of the project you want to fetch the customer-reports from
        :return: A Paginated response containing customer-reports
        """

        page = self._api.get(f"/customer-reports-service/api/v2/projects/{project_id}/customer-reports", **kwargs)
        return PaginatedResponse[CustomerReportEntity](**page)

    def get_one(self, customer_report_uri: str) -> CustomerReportEntity:
        """
        Get a single customer-report based on the customer-report uri

        :param customer_report_uri: the uri of the customer-report
        :return: A single customer-report or 404
        """
        resp = self._api.get(customer_report_uri)
        return CustomerReportEntity(**resp)

    def get_all_for_study(self, study_uri: str, **kwargs) -> PaginatedResponse[CustomerReportEntity]:
        """
        Get a single customer-report based on the customer-report uri

        :param study_uri: the uri of the study
        :return: A Paginated response containing customer-reports
        """
        study_uri = (study_uri
                     .replace("/studies-service/", "/customer-reports-service/")
                     .replace("/storage-service/", "/customer-reports-service/")
                     .replace("/v1/", "/v2/"))
        page = self._api.get(f"{study_uri}/customer-reports", **kwargs)
        return PaginatedResponse[CustomerReportEntity](**page)

    def download_customer_report_files(self, customer_report: CustomerReportEntity, out_path: str):
        """
        Download all files created by icobrain for a customer_report

        :param customer_report: The customer report you want to download the files from
        :param out_path: A folder path to write the files to
        :return:
        """
        if not os.path.isdir(out_path):
            os.mkdir(out_path)

        for report_file in customer_report.report_files:
            out_path_file = f"{out_path}/{report_file.name}"
            self.download_customer_report_file(report_file, out_path_file)

    def download_customer_report_file(self, customer_report_file: CustomerReportFile, out_path: Optional[str] = None):
        """
        Download a file created by icobrain for a customer_report

        :param customer_report_file: The file entity you want to download
        :param out_path: A path to write the file to
        :return:
        """
        if not out_path:
            out_path = customer_report_file.name
        logger.info(f"Downloading {customer_report_file} -> {out_path}")
        self._api.stream_file(customer_report_file.uri, out_path)

    def wait_for_customer_report_for_study(self, project_id: str,
                                           study_instance_uid: str,
                                           report_type: str) -> CustomerReportEntity:
        """
        After data has been imported a customer report will be created

        :param project_id: The ID of the project you want to search under
        :param study_instance_uid: The study instance UID from the DICOM
        :param report_type: The requested report type
        :return:
        """
        max_wait = 20  # sec (approx...)
        max_count = round(max_wait / self.polling_interval)
        count = 0
        while True:
            count += 1
            logger.info(f"Searching {report_type} report for {study_instance_uid} in {project_id}")
            params = {"study_instance_uid": study_instance_uid}
            for page in get_paginator(self.get_all, project_id=project_id, params=params):
                for report in page:
                    if report.icobrain_report_type == report_type:
                        return report
            if count >= max_count:
                raise IcometrixDataImportException(f"Failed to find a CustomerReport for {study_instance_uid}")
            sleep(self.polling_interval)

    def wait_for_results(self, customer_reports: List[CustomerReportEntity]) -> List[CustomerReportEntity]:
        """
        Wait until processing has finished and the result files are available on the customer report

        :param customer_reports: A list of customer reports
        :return:
        """
        customer_report_uris = [customer_report.uri for customer_report in customer_reports]

        finished_customer_reports: Dict[str, CustomerReportEntity] = {}
        while len(finished_customer_reports) != len(customer_reports):
            for customer_report_uri in customer_report_uris:
                if customer_report_uri in finished_customer_reports:
                    continue

                report = self.get_one(customer_report_uri)
                if report.status != "Finished":
                    logger.info(f"Waiting for {report}")
                    continue

                logger.info(f"Finished {report}")
                finished_customer_reports[report.uri] = report
            sleep(self.polling_interval)
        return [value for value in finished_customer_reports.values()]



import logging

from icometrix_sdk import IcometrixApi
from icometrix_sdk.utils.paginator import get_paginator

logging.basicConfig(level=logging.INFO)

STUDY_URI = "/storage-service/api/v1/projects/20c9e829-bc24-42bb-ac02-6527db61f4c5/patients/7d1cfc1e-a272-4c18-b24f-6f1a4393254a/studies/5fa896f5-37cd-48a7-9f2e-116453325bb2"

if __name__ == '__main__':
    ico_api = IcometrixApi()

    for cr_results in get_paginator(ico_api.customer_results.get_all_for_study, study_uri=STUDY_URI):
        for cr_result in cr_results:
            print(cr_result.uri)



import os
from io import BytesIO

import pydicom
from pydicom import Dataset
from pydicom.errors import InvalidDicomError

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.anonymizer.anonymizer import Anonymizer
from icometrix_sdk.anonymizer.policy import group_policy, policy_md5
from icometrix_sdk.models.upload_entity import StartUploadDto
from icometrix_sdk.anonymizer.hash_factory import HashFactory

PROJECT_ID = "uuid"
DICOM_DIR_PATH = "<path>"

# file_paths = ["IM-0007-0106.dcm"]
#
# hash_algo = HashFactory.create_hash_method("short_md5")
# anon = AnonymizerP(policy, group_policy, hash_algo)
#
# for file_path in file_paths:
#     dataset = pydicom.dcmread(file_path)
#     anon.anonymize(dataset).save_as(f"anon-{file_path}")

if __name__ == '__main__':
    hash_algo = HashFactory.create_hash_method("short_md5")
    anonymizer = Anonymizer(policy_md5, group_policy, hash_algo)

    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    # Get the project, to make sure its there (will throw a 404 in case the project is not found)
    project = ico_api.projects.get_one_by_id(PROJECT_ID)

    # prepare an upload entry, this entry can be used to push a block of files to (which is typical in DICOM)
    # Some options can be passed, for all available options please have a look at the StartUploadDto
    data = StartUploadDto(icobrain_report_type="icobrain_ms")
    upload = ico_api.uploads.start_upload(PROJECT_ID, data)

    # Loop over all files in a DIR and upload them to the before created upload entry
    for path, subdirs, files in os.walk(DICOM_DIR_PATH):
        for name in files:

            file_path = os.path.join(path, name)
            try:
                dicom_file = pydicom.dcmread(file_path)
            except InvalidDicomError:
                continue

            anonymized_dicom: Dataset = anonymizer.anonymize(dicom_file)
            buffer = BytesIO()
            anonymized_dicom.save_as(buffer)
            buffer.seek(0)
            fields = {"file": (file_path, buffer.read(), "application/octet-stream")}
            ico_api.uploads.upload_dicom(upload.uri, fields=fields)

    # Once all files have been uploaded, signal that they are all there and start the import/processing
    upload = ico_api.uploads.complete_upload(upload.uri)



from pathlib import Path

import pydicom
from pydicom.config import WARN
from pydicom.data import get_testdata_files

from icometrix_sdk.anonymizer.anonymizer import Anonymizer
from icometrix_sdk.anonymizer.hash_factory import HashFactory
from icometrix_sdk.anonymizer.policy import policy_sha, group_policy

# set env VALIDATION_MODE to 1 or 0, there are some invalid DICOMs in this set

# These files are included in the pydicom test dataset to test failed dcmread
INVALID_FILES = [
    "ExplVR_BigEndNoMeta.dcm",
    "ExplVR_LitEndNoMeta.dcm",
    "no_meta.dcm",
    "rtstruct.dcm",
    "OT-PAL-8-face.dcm",
]


def get_dicom_test_files():
    all_files = get_testdata_files("*.dcm")
    return [x for x in all_files if Path(x).name not in INVALID_FILES]


hash_algo = HashFactory.create_hash_method("md5")
anon = Anonymizer(policy_sha, group_policy, hash_algo)

for file_path in get_dicom_test_files():
    dataset = pydicom.dcmread(f"{file_path}")
    anon.anonymize(dataset).save_as(f"out/anon-{Path(file_path).name}")



import hashlib
from abc import abstractmethod
from typing import get_args

from icometrix_sdk.anonymizer.exceptions import HashAlgorithmException, HashSizeException
from icometrix_sdk.anonymizer.models import HashAlgo


class HashFactory:
    @staticmethod
    def create_hash_method(algo: HashAlgo, size=512, salt=None):
        if algo == "sha3":
            return SHA3(size)
        elif algo == "md5":
            return MD5()
        elif algo == "short_md5":
            return ShortMD5()
        else:
            supported = ", ".join(get_args(HashAlgo))
            raise HashAlgorithmException(f"No algorithm named {algo} is supported, valid values are {supported}")


class HashMethod:
    @abstractmethod
    def calculate_hash_from_bytes(self, input_obj: bytes) -> str:
        pass

    def calculate_hash(self, input_obj: str, encoding='utf-8') -> str:
        return self.calculate_hash_from_bytes(input_obj.encode(encoding))


class SHA3(HashMethod):
    def __init__(self, size=256):
        acceptable_capacities = [224, 256, 384, 512]
        if size not in acceptable_capacities:
            raise f"Invalid capacity for SHA3, should be any of {acceptable_capacities}"
        self.size = size

    def calculate_hash_from_bytes(self, input_obj: bytes):
        if self.size == 224:
            hash_obj = hashlib.sha3_224(input_obj)
        elif self.size == 256:
            hash_obj = hashlib.sha3_256(input_obj)
        elif self.size == 384:
            hash_obj = hashlib.sha3_384(input_obj)
        elif self.size == 512:
            hash_obj = hashlib.sha3_512(input_obj)
        else:
            raise HashSizeException(f"SHA3 does not support size {self.size}, "
                                    f"valid values are (224, 256, 384, 512)")

        digest = hash_obj.hexdigest()
        return digest


class MD5(HashMethod):
    def calculate_hash_from_bytes(self, input_obj: bytes):
        return hashlib.md5(input_obj, usedforsecurity=True).hexdigest()


class ShortMD5(HashMethod):
    """
    MD5 that is re-based to base10.
    """

    def calculate_hash(self, input_obj: str, encoding='utf-8') -> str:
        return self.calculate_hash_from_bytes(input_obj.encode(encoding))

    def calculate_hash_from_bytes(self, input_obj: bytes):
        md5_hash = MD5().calculate_hash_from_bytes(input_obj)
        return str(int(md5_hash, base=16))[:10]



from typing import Optional, Literal, Union, Any

from icometrix_sdk.models.base import BackendEntity

PipelineResultType = Literal[
    "unusable_finalization",
    "icobrain_mr_cross",
    "icobrain_ctp_cross",
    "icobrain_mt",
    "icobrain_ct_cross",
    "protocol_validation",
    "icobrain_preprocessing",
    "harmonization_constants",
    "icobrain_diffusion",
    "pdf2dicom",
    "icolung",
    "icobrain_mr_cross_research",
    "icobrain_mr_cross_manual_corrections",
    "icospine_cross",
    "icobrain_mt_cross",
]
PipelineResultJobType = Literal[
    "overread",
    "corrected",
    "curated",
    "initial",
]


class PipelineResultEntity(BackendEntity):
    image: str
    job_id: str
    results: Any
    job_mode: Optional[Union[PipelineResultJobType, str]] = None
    pipeline: Union[PipelineResultType, str]
    study_id: str
    patient_id: str
    project_id: str
    run_parameters: dict
    automatic_qc_results: Optional[dict] = None
    software_version_data: dict



import pytest
from pydicom import DataElement
from pydicom.valuerep import MAX_VALUE_LEN

from icometrix_sdk.anonymizer.hash_factory import HashFactory, HashMethod
from icometrix_sdk.anonymizer.utils import _cut_max_length
from icometrix_sdk.anonymizer.utils import hash_tag


@pytest.fixture(scope="module")
def hash_algo() -> HashMethod:
    return HashFactory.create_hash_method("md5")


def test_hash_long_value(hash_algo: HashMethod):
    value = "Head"
    max_len = MAX_VALUE_LEN["LO"]

    expected = _cut_max_length(hash_algo.calculate_hash(value), max_len)

    elem1 = DataElement(0x00080013, "LO", value)
    hash_tag(elem1, hash_algo)

    assert elem1.value == expected


def test_hash_short_value(hash_algo: HashMethod):
    value = "4753014.1"
    max_len = MAX_VALUE_LEN["SH"]

    expected = _cut_max_length(hash_algo.calculate_hash(value), max_len)

    elem1 = DataElement(0x00080050, "SH", value)
    hash_tag(elem1, hash_algo)

    assert elem1.value == expected


def test_hash_number_value(hash_algo: HashMethod):
    elem1 = DataElement(0x00020000, "UL", 210)
    with pytest.raises(ValueError):
        hash_tag(elem1, hash_algo)


def test_hash_ui_value(hash_algo: HashMethod):
    value = "1.2.826.0.1.3680043.9.5542.5114248473116471214116117310121520961"
    expected = "1.2.826.0.1.3680043.9.5542.2676173402192025550109336474686546713"
    elem1 = DataElement(0x0020000D, "UI", value)
    hash_tag(elem1, hash_algo)
    assert elem1.value == expected



from inspect import signature
from typing import Callable, Generic, TypeVar, Optional, Iterable, ParamSpec

from icometrix_sdk.models.base import PaginatedResponse

T = TypeVar("T")
P = ParamSpec("P")


def can_paginate(func: Callable):
    """
    Checks if a function is iterable
    """
    sig = signature(func)
    return issubclass(sig.return_annotation, PaginatedResponse)


class PageIterator(Generic[T], Iterable):
    """
    An iterable object to iterate over paginated api responses
    """

    _current_page: Optional[PaginatedResponse[T]] = None

    def __init__(self, func: Callable, op_kwargs, page_size: int = 50, starting_index=0):
        self._func = func
        self._op_kwargs = op_kwargs
        self._page_size = page_size
        self._starting_index = starting_index
        self._page_index = starting_index
        self._params = {"pageIndex": self._page_index, "pageSize": self._page_size}

    def __iter__(self):
        self._page_index = self._starting_index
        self._params = {"pageIndex": self._page_index, "pageSize": self._page_size}
        return self

    def __next__(self) -> PaginatedResponse[T]:
        if self._current_page is None:
            self._current_page = self._fetch_current_page()
            self._page_index = self._page_index + 1
            return self._current_page
        if not self._current_page.has_next():
            raise StopIteration

        self._current_page = self._fetch_current_page()
        self._page_index = self._page_index + 1
        return self._current_page

    def _fetch_current_page(self) -> PaginatedResponse[T]:
        self._params = {"pageIndex": self._page_index, "pageSize": self._page_size}
        if "params" in self._op_kwargs:
            self._op_kwargs["params"].update(self._params)
        else:
            self._op_kwargs["params"] = self._params
        return self._func(**self._op_kwargs)


def get_paginator(func: Callable[..., PaginatedResponse[T]],
                  page_size: Optional[int] = 50,
                  starting_index: Optional[int] = 0, **kwargs) -> PageIterator[T]:
    """
    Create paginator object for an operation.

    This returns an iterable object. Iterating over
    this object will yield a single page of a response
    at a time.

    :param func:
        The function that needs to be paginated. (The function needs
         to return a :class:`~icometrix_sdk.models.base.PaginatedResponse`)
    :param page_size:
        The size of the pages
    :param starting_index:
        The starting page
    :returns: A PageIterator
    """
    if not can_paginate(func):
        raise ValueError(f"Function '{func.__name__}' can't be paginated")

    # sig = signature(func)
    return PageIterator[T](func, op_kwargs=kwargs, page_size=page_size,
                           starting_index=starting_index)



from copy import deepcopy

import pydicom
import pytest
from pydicom import Dataset
from pydicom.data import get_testdata_file

from icometrix_sdk.anonymizer.anonymizer import Anonymizer
from icometrix_sdk.anonymizer.hash_factory import HashMethod, HashFactory
from icometrix_sdk.anonymizer.models import Policy, TagPolicy
from icometrix_sdk.anonymizer.tests.utils import _ignore_tag, replace_by_value


@pytest.fixture(scope="module")
def hash_algo() -> HashMethod:
    return HashFactory.create_hash_method("md5")


@pytest.fixture(scope="function")
def test_dataset() -> Dataset:
    dataset = pydicom.read_file(get_testdata_file("MR_small.dcm"))
    dataset.PatientBirthDate = "19930831"
    return dataset


remove_all: Policy = {}


def test_remove_all_policy(hash_algo: HashMethod, test_dataset: Dataset):
    anon = Anonymizer(remove_all, remove_all, hash_algo)
    anon.anonymize(test_dataset)
    for element in test_dataset:
        if _ignore_tag(element.tag):
            continue

        assert element.value in (0, 0.0, "")


keep_values: Policy = {
    0x00180081: TagPolicy("keep", "EchoTime"),
    0x00100040: TagPolicy("keep", "PatientsSex"),
}


def test_keep_policy(hash_algo: HashMethod, test_dataset: Dataset):
    original = deepcopy(test_dataset)
    anon = Anonymizer(keep_values, {}, hash_algo)
    anon.anonymize(test_dataset)
    for tag in keep_values:
        assert test_dataset[tag].value == original[tag].value


hash_values: Policy = {
    0x00080018: TagPolicy("hash", "SOPInstanceUID"),
    0x0020000e: TagPolicy("hash", "SeriesInstanceUID"),
    0x00100020: TagPolicy("hash", "PatientId"),
}


def test_hash_policy(hash_algo: HashMethod, test_dataset: Dataset):
    original = deepcopy(test_dataset)
    anon = Anonymizer(hash_values, {}, hash_algo)
    anon.anonymize(test_dataset)
    for tag in hash_values:
        assert test_dataset[tag].value != original[tag].value


replace_values: Policy = {
    0x00100010: TagPolicy("replace", "PatientName", replace_by_value("Frodo")),
    0x00080080: TagPolicy("replace", "InstitutionName", replace_by_value("Mordor")),
}


def test_replace_policy(hash_algo: HashMethod, test_dataset: Dataset):
    anon = Anonymizer(replace_values, {}, hash_algo)
    anon.anonymize(test_dataset)

    test_dataset[0x00100010].value = "Frodo"
    test_dataset[0x00080080].value = "Mordor"


def test_invalid_replace_policy(hash_algo: HashMethod, test_dataset: Dataset):
    with pytest.raises(ValueError):
        anon2 = Anonymizer({0x00080080: TagPolicy("replace", "InstitutionName")}, {}, hash_algo)
        anon2.anonymize(test_dataset)


round_values: Policy = {
    0x00100030: TagPolicy("round", "PatientBirthDate"),
}


def test_round_policy(hash_algo: HashMethod, test_dataset: Dataset):
    anon = Anonymizer(round_values, {}, hash_algo)
    anon.anonymize(test_dataset)
    for tag in round_values:
        assert test_dataset[tag].value.endswith("0101")



import json
import logging
import os
import shutil
import uuid
from typing import Optional, Type
from urllib.parse import urljoin

import requests
from requests import Response, HTTPError, Session
from requests.adapters import HTTPAdapter
from urllib3.util import Retry

from icometrix_sdk._version import __version__
from icometrix_sdk.authentication import AuthenticationMethod
from icometrix_sdk.exceptions import IcometrixHttpException, IcometrixParseException, IcometrixConfigException, \
    IcometrixAuthException
from icometrix_sdk.logger import logger_name
from icometrix_sdk.utils.api_client import ApiClient
from icometrix_sdk.utils.file_upload import create_multipart

HTTP_TIMEOUT = os.getenv("HTTP_TIMEOUT", 120)

logger = logging.getLogger(logger_name)


class RequestsApiClient(ApiClient):
    """
    An implementation of the API client using the requests library

    Dev Note: Try not the expose requests logic, we might want to swap this later on
    """

    _auth_attempts = 0

    _session: Session
    auth: Optional[AuthenticationMethod]

    def __init__(self, server: str, auth: Optional[AuthenticationMethod] = None):
        self.base_headers = {
            "User-Agent": "Python SDK v{}".format(__version__),
            "x-sdk-type": "Python",
            "x-sdk-version": __version__,
            "x-sdk-contextId": str(uuid.uuid4()),
            "Accept": "application/json",
            "Content-Type": "application/json"
        }
        self.server = server
        self.auth = auth

        if not server:
            raise IcometrixConfigException("Server is required")

        self._session = self._create_session()
        self._authenticate()

    def _authenticate(self):
        if self.auth:
            logger.info("Authenticating")
            self._auth_attempts += 1
            self.auth.connect(self)
            self._auth_attempts = 0

    def _refresh_token(self, resp: Response):
        if self._auth_attempts > 0:
            raise IcometrixAuthException("Failed to authenticate")
        if resp.status_code == 401:
            logger.info("Fetching new token as the previous token expired")
            self._authenticate()

    def _create_session(self) -> Session:
        """
        Create a new session
        """
        session = requests.Session()
        retries = Retry(
            total=3,
            backoff_factor=0.5,
            status_forcelist=[429, 502, 503, 504],
        )
        session.mount('https://', HTTPAdapter(max_retries=retries))
        session.headers.update(self.base_headers)
        return session

    def _make_request(self,
                      method=Type[str],
                      url=Type[str],
                      data: Optional[any] = None,
                      headers: Optional[dict] = None,
                      params: Optional[dict] = None,
                      files=None,
                      cookies=None,
                      stream=False) -> Response:
        """
        private function to abstract most of the HTTP request logic

        :param method: GET, POST, PUT, PATCH, DELETE...
        :param url: A relative URL to specify the API endpoint
        :param data: Dict containing data for the request body
        :param headers: Dict containing additional request headers.
        :param params: Dict containing query parameters.
        :param files: (optional) Dict of ``'filename': file-like-objects`` for multipart encoding upload.
        :param cookies: (optional) Dict or CookieJar object to send with the
        :return:
        """

        full_url = urljoin(self.server, url)
        try:
            response = self._session.request(
                method=method,
                url=full_url,
                data=data,
                headers=headers,
                params=params,
                files=files,
                timeout=HTTP_TIMEOUT,
                cookies=cookies,
                stream=stream
            )
            response.raise_for_status()
        except HTTPError as e:
            message = "Exception received when sending an HTTP request"
            logger.exception(message)
            raise IcometrixHttpException(e)

        return response

    def get(self, uri: str, **kwargs) -> dict:
        """
        Submit a GET request to the API.

        :param uri: A relative URL to specify the API endpoint
        :param params: A dictionary containing query parameters.
        :returns: The API response as a dictionary.
        """

        resp = self._make_request(
            method="GET",
            url=uri,
            **kwargs
        )
        return response_to_dict(resp)

    def post(self, uri: str, data: dict, **kwargs) -> dict:
        """
        Submit a POST request to the API.

        :param uri: A relative URL to specify the API endpoint
        :param data: A dictionary containing data for the request body
        :returns: The API response as a dictionary.

        """

        resp = self._make_request(
            method="POST",
            url=uri,
            data=json.dumps(data),
            **kwargs
        )
        return response_to_dict(resp)

    def put(self, uri: str, data: dict, **kwargs) -> dict:
        """
        Submit a PUT request to the API.

        :param uri: A relative URL to specify the API endpoint
        :param data: A dictionary containing data for the request body
        :returns: The API response as a dictionary.

        """

        resp = self._make_request(
            method="PUT",
            url=uri,
            data=json.dumps(data),
            **kwargs
        )
        return response_to_dict(resp)

    def delete(self, uri: str, **kwargs):
        """
        Submit a DELETE request to the API.

        :param uri: A relative URL to specify the API endpoint
        :returns: void

        """
        self._make_request(
            method="DELETE",
            url=uri,
            **kwargs
        )

    def put_file(self, uri: str, fields, **kwargs):
        """
        Submit a PUT request to the API.

        :param uri: A relative URL to specify the API endpoint
        :param fields: Dictionary of fields or list of (key, :class:`~urllib3.fields.RequestField`).
        :returns:

        """

        data, headers = create_multipart(fields)
        self._make_request(
            method="PUT",
            url=uri,
            data=data,
            headers=headers,
            **kwargs
        )

    def stream_file(self, uri: str, out_path: str, **kwargs):
        with self._make_request('GET', uri, stream=True) as r:
            with open(out_path, 'wb') as f:
                shutil.copyfileobj(r.raw, f)


def response_to_dict(response: Response) -> dict:
    content = response.content
    try:
        return json.loads(content)
    except ValueError as e:
        raise IcometrixParseException("Failed to parse response to json: {}".format(e.args[0]))



from typing import Optional

from pydantic import field_validator

from icometrix_sdk.exceptions import IcometrixInvalidInputDataException
from icometrix_sdk.models.base import BackendEntity, DicomModality, utc_datetime


class StudyEntity(BackendEntity):
    modality: DicomModality
    patient_id: str
    project_id: str

    study_date: str  # DICOM formatted study date, e.g. 20100513
    study_note: Optional[str] = None
    study_time: str  # DICOM formatted study time
    patient_age: Optional[int | str] = None  # Can be something like 035Y
    manufacturer: Optional[str] = None
    accession_number: Optional[str] = None
    institution_name: Optional[str] = None
    study_description: Optional[str] = None
    study_instance_uid: Optional[str] = None
    scheduled_timestamp: Optional[str] = None
    device_serial_number: Optional[str] = None
    icobrain_report_type: Optional[str] = None
    magnetic_field_strength: Optional[str] = None
    manufacturer_model_name: Optional[str] = None
    imported_timestamp: Optional[utc_datetime] = None

    @field_validator("study_date")
    def is_valid_study_date(cls, value):
        if value.isdigit() and int(value) > 0 and len(value) == 8:
            return value
        else:
            raise IcometrixInvalidInputDataException(
                "Study date does not follow DICOM format (e.g. 20220117)")



from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.patient_entity import PatientEntity
from icometrix_sdk.utils.requests_api_client import ApiClient


class Patients:
    def __init__(self, api: ApiClient):
        self._api = api

    def get_all(self, project_uri: str, **kwargs) -> PaginatedResponse[PatientEntity]:
        """
        List all patients within a project

        :param: The uri of the project
        :return: A Paginated response containing patients
        """
        page = self._api.get(f"{project_uri}/patients", **kwargs)
        return PaginatedResponse[PatientEntity](**page)

    def get_one(self, patient_uri: str) -> PatientEntity:
        """
        Get a single patient based on the patient uri

        :param patient_uri: the uri of the patient
        :return: A single patient or 404
        """
        resp = self._api.get(patient_uri)
        return PatientEntity(**resp)



class IcometrixException(Exception):
    ...


class PolicyException(IcometrixException):
    ...


class HashAlgorithmException(IcometrixException):
    ...


class HashSizeException(IcometrixException):
    ...



import os
from abc import ABC, abstractmethod
from typing import Optional

from icometrix_sdk.exceptions import IcometrixConfigException
from icometrix_sdk.utils.api_client import ApiClient

TOKEN_URI = "/authentication-service/api/v1/token-sessions"
SESSION_URI = "/authentication-service/api/v1/sessions"


class AuthenticationMethod(ABC):
    """
    Interface for authentication methods
    """

    @abstractmethod
    def connect(self, api: ApiClient):
        pass

    @abstractmethod
    def disconnect(self, api: ApiClient):
        pass


class PasswordAuthentication(AuthenticationMethod):
    """
    Email/Password authentication method
    """

    def __init__(self, email: str, password: str):
        self.email = email
        self.password = password

    def connect(self, api: ApiClient):
        data = {"email": self.email, "password": self.password}
        api.post(SESSION_URI, data=data)

    def disconnect(self, api: ApiClient):
        if api:
            api.delete(SESSION_URI)


class TokenAuthentication(AuthenticationMethod):
    """
    Token authentication method
    """

    def __init__(self, token: str):
        self.token = token

    def connect(self, api: ApiClient):
        data = {"key": self.token}
        api.post(TOKEN_URI, data=data)

    def disconnect(self, api: ApiClient):
        if api:
            api.delete(SESSION_URI)


def get_auth_method() -> Optional[AuthenticationMethod]:
    """
    Find an authentication method based on the 'AUTH_METHOD' environment variable
    """
    auth_method = os.getenv("AUTH_METHOD")
    if auth_method is None:
        if os.getenv("TOKEN"):
            auth_method = "token"
        elif os.getenv("EMAIL") and os.getenv("PASSWORD"):
            auth_method = "basic"

    if auth_method == "basic":
        return PasswordAuthentication(os.environ["EMAIL"], os.environ["PASSWORD"])
    elif auth_method == "token":
        return TokenAuthentication(os.environ["TOKEN"])

    raise IcometrixConfigException("Please set an 'AUTH_METHOD' (basic OR token)")



import logging
import os

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.models.upload_entity import StartUploadDto

PROJECT_ID = "pid"

# A directory containing a T1 and FLAIR scan, as required for icobrain_ms for more info see HCP manual
DICOM_DIR_PATH = "..."

logging.basicConfig(level=logging.INFO)

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    # Initialize the icometrix API
    ico_api = IcometrixApi()

    # Get the project, to make sure its there (will throw a 404 in case the project is not found)
    project = ico_api.projects.get_one_by_id(PROJECT_ID)

    # Upload a directory of DICOMs
    data = StartUploadDto(icobrain_report_type="icobrain_ms")
    upload = ico_api.uploads.start_upload(PROJECT_ID, data)
    count = ico_api.uploads.upload_all_files_in_dir(upload.uri, DICOM_DIR_PATH)

    # add validation logic here
    file_uploads = ico_api.uploads.get_uploaded_files(upload.folder_uri)
    if len(file_uploads.files) != count:
        raise Exception("Not all files uploaded")

    # Once all files have been uploaded, signal that they are all there and start the import/processing
    upload = ico_api.uploads.complete_upload(upload.uri)



import os

from pydicom.config import RAISE

# Organization root, default to icometrix (https://dicom.nema.org/dicom/2013/output/chtml/part05/chapter_B.html)
ROOT_UID: str = os.getenv("ROOT_UID", "1.2.826.0.1.3680043.9.5542")
VALIDATION_MODE: int = int(os.getenv("VALIDATION_MODE", RAISE))

PATIENT_IDENTITY_REMOVED_TAG: int = 0x00120062
DE_IDENTIFICATION_METHOD_TAG: int = 0x00120063
PRIVATE_ICOMETRIX_GROUPS: list[int] = [0x0009, 0x0015, 0x0017]



import copy

import pydicom
import pytest

from icometrix_sdk.anonymizer.anonymizer import Anonymizer
from icometrix_sdk.anonymizer.exceptions import PolicyException
from icometrix_sdk.anonymizer.hash_factory import HashMethod, HashFactory
from icometrix_sdk.anonymizer.models import Policy, TagPolicy
from icometrix_sdk.anonymizer.policy import policy_sha, policy_md5, group_policy
from icometrix_sdk.anonymizer.tests.utils import _ignore_tag


@pytest.fixture(scope="module")
def hash_algo() -> HashMethod:
    return HashFactory.create_hash_method("md5")


tags: Policy = {
    0x00180081: TagPolicy("keep", "EchoTime"),
}

groups: Policy = {
    0x0018: TagPolicy("keep", "Group 18"),
}


def test_valid_constructor(hash_algo: HashMethod):
    Anonymizer(tags, groups, hash_algo)


def test_invalid_constructor(hash_algo: HashMethod):
    with pytest.raises(PolicyException):
        Anonymizer(groups, groups, hash_algo)

    with pytest.raises(PolicyException):
        Anonymizer(tags, tags, hash_algo)


def test_full_md5(hash_algo: HashMethod):
    anonymizer = Anonymizer(policy_md5, group_policy, hash_algo)

    original_ds = pydicom.dcmread(f"datasets/original.dcm")
    expected_ds = pydicom.dcmread(f"datasets/anonymized_md5.dcm")

    anonymized_ds = anonymizer.anonymize(copy.deepcopy(original_ds))

    for element in anonymized_ds:
        tag = element.tag
        if element.VR == "SQ" or _ignore_tag(tag):
            continue

        assert anonymized_ds[tag].value == expected_ds[tag].value


def test_full_sha3(hash_algo: HashMethod):
    anonymizer = Anonymizer(policy_sha, group_policy, hash_algo)

    original_ds = pydicom.dcmread(f"datasets/original.dcm")
    expected_ds = pydicom.dcmread(f"datasets/anonymized_sha3.dcm")

    anonymized_ds = anonymizer.anonymize(copy.deepcopy(original_ds))

    for element in anonymized_ds:
        tag = element.tag
        if element.VR == "SQ" or _ignore_tag(tag):
            continue

        assert anonymized_ds[tag].value == expected_ds[tag].value



from pydantic import BaseModel
from typing_extensions import Optional

from icometrix_sdk.models.base import BackendEntity


class CustomerResultPipelineResults(BaseModel):
    pipeline_results_id: Optional[str] = None


class CustomerResultEntity(BackendEntity):
    uri: str
    job_id: str
    study_id: str
    patient_id: str
    project_id: str
    qc_result_id: str
    pipeline_results: CustomerResultPipelineResults



from icometrix_sdk.models.user_entity import User
from icometrix_sdk.utils.api_client import ApiClient


class Profile:
    def __init__(self, api: ApiClient):
        self._api = api

    def who_am_i(self) -> User:
        """
        see the currently logged-in user

        :return: The current User or 401
        """
        resp = self._api.get(f"/authentication-service/api/v1/loginstatus")
        return User(**resp)



from time import sleep
from typing import Callable, TypeVar

from icometrix_sdk.exceptions import WaitTimeoutException

T = TypeVar('T')


def wait_for(func: Callable[..., T], condition: Callable[[T], bool], timeout: int, **kwargs) -> T:
    pol = 5
    duration = 0
    val = func(**kwargs)
    while not condition(val):
        if timeout >= duration:
            WaitTimeoutException("The function took to long to complete")
        duration += pol
        sleep(pol)
        val = func(**kwargs)
    return val



from copy import deepcopy

import pydicom
import pytest
from pydicom import Dataset
from pydicom.data import get_testdata_file

from icometrix_sdk.anonymizer.anonymizer import Anonymizer
from icometrix_sdk.anonymizer.hash_factory import HashMethod, HashFactory
from icometrix_sdk.anonymizer.models import Policy, TagPolicy
from icometrix_sdk.anonymizer.tests.utils import _ignore_tag, _split_group_tags


@pytest.fixture(scope="module")
def hash_algo() -> HashMethod:
    return HashFactory.create_hash_method("md5")


@pytest.fixture(scope="function")
def test_dataset() -> Dataset:
    dataset = pydicom.read_file(get_testdata_file("MR_small.dcm"))
    dataset.PatientBirthDate = "19930831"
    return dataset


keep_group: Policy = {
    0x0018: TagPolicy("keep", "Group18"),
}


def test_keep_group_policy(hash_algo: HashMethod, test_dataset: Dataset):
    original = deepcopy(test_dataset)
    anon = Anonymizer({}, keep_group, hash_algo)
    anon.anonymize(test_dataset)
    in_group, outside_group = _split_group_tags(test_dataset, 0x0018)
    for element in in_group:
        assert element.value == original[element.tag].value

    for element in outside_group:
        if _ignore_tag(element.tag):
            continue
        assert element.value in (0, 0.0, "")


hash_group: Policy = {
    0x0008: TagPolicy("hash", "HashGroup"),
}


def test_hash_policy(hash_algo: HashMethod, test_dataset: Dataset):
    original = deepcopy(test_dataset)
    anon = Anonymizer({}, hash_group, hash_algo)
    anon.anonymize(test_dataset)
    in_group, outside_group = _split_group_tags(test_dataset, 0x0008)

    un_hashable_vrs = ["CS", "DA", "TM"]
    for element in in_group:
        if not element.value or element.VR in un_hashable_vrs:
            continue
        assert element.value != original[element.tag].value



import logging
import os

from icometrix_sdk import IcometrixApi, Region
from icometrix_sdk.utils.paginator import get_paginator

logging.basicConfig(level=logging.INFO)

if __name__ == '__main__':

    os.environ["API_HOST"] = Region.EU.value

    ico_api = IcometrixApi()

    for projects in get_paginator(ico_api.projects.get_all, page_size=5):
        for project in projects:
            print(project.name)



from datetime import datetime, timedelta, timezone


def time_difference(utc_datetime: datetime) -> timedelta:
    """
    Parses the input datetime string and logs the time difference with the current time.

    :param utc_datetime: The datetime to be compared

    Returns:
    """
    utc_datetime = utc_datetime.replace(tzinfo=timezone.utc)
    current_utc_datetime = datetime.now(timezone.utc)
    return current_utc_datetime - utc_datetime



"""
An example on how to do authentication based on an 'AUTH_METHOD' set as environment variable
"""
import os

from icometrix_sdk import IcometrixApi

if __name__ == '__main__':
    os.environ["API_HOST"] = Region.EU.value

    ico_api = IcometrixApi()

    print(ico_api.profile.who_am_i().email)



from icometrix_sdk.anonymizer.utils import _cut_max_length, is_group, is_tag


def test_cut_value():
    max_length = 4
    value = "123456789"
    expected = "1234"
    assert _cut_max_length(value, max_length) == expected


def test_cut_short_value():
    max_length = 20
    value = "123456789"
    assert _cut_max_length(value, max_length) == value


def test_is_group():
    assert is_group(0x0000) is True
    assert is_group(0x0010) is True
    assert is_group(0x0018) is True
    assert is_group(0xffff) is True

    assert is_group(0x00100000) is False
    assert is_group(0x0018002c) is False
    assert is_group(0xffffffff) is False


def test_is_tag():
    assert is_tag(0x0000) is False
    assert is_tag(0x0010) is False
    assert is_tag(0x0018) is False
    assert is_tag(0xffff) is False

    assert is_tag(0x00100000) is True
    assert is_tag(0x0018002c) is True
    assert is_tag(0xffffffff) is True



from pydicom import DataElement, Dataset
from pydicom.valuerep import INT_VR, FLOAT_VR, validate_value, MAX_VALUE_LEN

from icometrix_sdk.anonymizer.config import ROOT_UID, PATIENT_IDENTITY_REMOVED_TAG, DE_IDENTIFICATION_METHOD_TAG, \
    VALIDATION_MODE
from icometrix_sdk.anonymizer.hash_factory import ShortMD5, SHA3
from icometrix_sdk.anonymizer.models import ReplaceFn


def is_group(tag: int) -> bool:
    # Will fail on 0x00000000, but this does not exist in DICOM...
    return (tag >> 16) == 0


def is_tag(tag: int) -> bool:
    # Same as is group will fail on 0x00000000, but this does not exist in DICOM...
    return (tag >> 16) != 0


def _is_pixel_data(tag: int) -> bool:
    return tag == 0x7fe00010


def _is_numeric_vr(vr: str) -> bool:
    return vr in FLOAT_VR or vr in INT_VR


def empty_tag(element: DataElement):
    """
    Empties the value of the DICOM element.
    For numeric VRs, sets the value to 0.
    For other VRs, sets the value to an empty string.
    """

    if _is_numeric_vr(element.VR):
        element.value = 0
    else:
        element.value = ""


def remove_tag(element: DataElement, dataset: Dataset):
    """
    Removes the tag from a dataset
    """
    tag = element.tag
    if tag in dataset:
        del dataset[tag]


def replace_tag(element: DataElement, dataset: Dataset, replace_fn: ReplaceFn):
    """
    Replaces the value of the DICOM element with a new value.
    Raises an ValueError Exception if the new value is not compatible with the VR of the DICOM element.

    value: The new value to replace the current value of the DICOM element.
    """
    if replace_fn is None:
        raise ValueError("No Replace function provided to the policy")
    replace_fn(element, dataset)
    validate_value(element.VR, element.value, validation_mode=VALIDATION_MODE)


def _hash_uid(element: DataElement, hash_method):
    """
    Calculates the hash of the value of the DICOM UID element
    """
    max_len = 64
    hashed = hash_method.calculate_hash(element.value)
    # Convert hash in hexadecimal format to decimal format
    # Requirement of DICOM UIDs that they only exist of digits
    # See https://dicom.nema.org/dicom/2013/output/chtml/part05/chapter_9.html
    hashed = str(int(hashed, base=16))

    # Dicom 9.1: The first digit of each component shall not be zero,
    # unless the component is a single digit.
    # See https://dicom.nema.org/dicom/2013/output/chtml/part05/chapter_9.html
    extra = ""
    if hashed[0] == "0":
        extra = "9"
    new_value = f"{ROOT_UID}.{extra}{hashed}"

    # DICOM 9.1: UIDs, shall not exceed 64 total characters
    return _cut_max_length(new_value, max_len)


def _cut_max_length(value: str, max_len: int) -> str:
    if max_len and len(value) > max_len:
        value = value[:max_len]
    return value


def hash_tag(element: DataElement, hash_method):
    """
    Calculates the hash of the value of the DICOM element
    """

    if _is_numeric_vr(element.VR):
        raise ValueError(f"Cant hash VR {element.VR}")

    if element.VR == "UI":
        hashed = _hash_uid(element, hash_method)
    else:
        hashed = hash_method.calculate_hash(element.value)
        if element.VR in MAX_VALUE_LEN:
            hashed = _cut_max_length(hashed, MAX_VALUE_LEN[element.VR])

    validate_value(element.VR, hashed, validation_mode=VALIDATION_MODE)
    element.value = hashed


def _round_da(element: DataElement):
    """
    The structure of a DA is YYYYMMDD
    This will be rounded to YYYY0101
    """
    if len(element.value) != 8:
        return
    element.value = element.value[:4] + "0101"


def _round_dt(element: DataElement):
    """
    The structure of a DT is  YYYYMMDDHHMMSS.FFFFFF&ZZXX (&ZZXX is an optional suffix for offset from UTC)
    This will be rounded to YYYY01010000.000000
    """
    if not element.value:
        return
    element.value = element.value[:4] + "0101000000.000000"


def _round_tm(element: DataElement):
    """
    The structure of a DT is  YYYYMMDDHHMMSS.FFFFFF&ZZXX (&ZZXX is an optional suffix for offset from UTC)
    This will be rounded to YYYY01010000.000000
    """
    if not element.value:
        return
    element.value = "000001.00000"


def round_tag(element: DataElement):
    if element.VR == "DA":
        _round_da(element)
    elif element.VR == "DT":
        _round_dt(element)
    elif element.VR == "TM":
        _round_tm(element)
    else:
        raise ValueError(f"Cant round VR {element.VR}")


def add_de_identification_tags(dataset: Dataset) -> Dataset:
    # Field "PatientIdentityRemoved"
    # See https://dicom.innolitics.com/ciods/rt-plan/patient/00120062
    dataset.add_new(PATIENT_IDENTITY_REMOVED_TAG, "LT", "YES")

    # Field "De-identification method"
    # See https://dicom.innolitics.com/ciods/rt-plan/patient/00120063
    dataset.add_new(DE_IDENTIFICATION_METHOD_TAG, "LT",
                    "Python based, HIPAA compliant, based on DICOM PS3.15 AnnexE.")
    return dataset


def short_md5_hash(element: DataElement, _):
    value = element.value
    if len(value) % 2 != 0:
        value += " "
    element.value = ShortMD5().calculate_hash(value)


def short_sha3_hash(element: DataElement, _):
    value = str(element.value)
    if len(value) % 2 != 0:
        value += " "
    element.value = SHA3(size=512).calculate_hash(value)[:10]


def remove_if_birthday(element: DataElement, ds: Dataset):
    if ds[0x00100030].value:
        del ds[element.tag]



import pytest
from pydicom import DataElement

from icometrix_sdk.anonymizer.utils import round_tag


# DA VR
def test_round_da_value():
    expected = "20180101"
    elem1 = DataElement(0x00080012, "DA", "20180629")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_empty_da_value():
    expected = ""
    elem1 = DataElement(0x00080012, "DA", "")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_invalid_short_da_value():
    expected = "1720"
    elem1 = DataElement(0x00080012, "DA", "1720")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_invalid_long_da_value():
    expected = "2500445566"
    elem1 = DataElement(0x00080012, "DA", "2500445566")
    round_tag(elem1)

    assert elem1.value == expected


# DT VR
def test_round_dt_value():
    expected = "20180101000000.000000"
    elem1 = DataElement(0x0008002A, "DT", "20180629072603.066406")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_empty_dt_value():
    expected = ""
    elem1 = DataElement(0x00080012, "DT", "")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_invalid_short_dt_value():
    expected = "17200101000000.000000"
    elem1 = DataElement(0x00080012, "DT", "17200204")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_invalid_long_dt_value():
    expected = "20180101000000.000000"
    elem1 = DataElement(0x00080012, "DT", "20180629072603.066406984135")
    round_tag(elem1)

    assert elem1.value == expected


# Numbers...
def test_round_int_value():
    expected = "20180101"
    elem1 = DataElement(0x00080012, "DA", "20180629")
    round_tag(elem1)

    assert elem1.value == expected


def test_round_unsupported_vr():
    elem1 = DataElement(0x00081000, "LO", "Model")
    with pytest.raises(ValueError):
        round_tag(elem1)



from icometrix_sdk.anonymizer.hash_factory import HashFactory


def test_hash_string_hash_factory():
    # Input data
    data = 'Patient Name'

    # Expected SHA3-256 hash
    expected_hash = '95fa47c8fb19996e4bdeed2c3fa7cc35ab013737fd8b562a51335302773c4b29'

    # Calculate the SHA3-256 hash
    hash_factory = HashFactory()
    hash_method = hash_factory.create_hash_method("sha3", 256)
    hash_digest = hash_method.calculate_hash(data)  # anonymizer._hash(data)

    # Verify the calculated hash matches the expected hash
    assert expected_hash == hash_digest


def test_sha3():
    dataset = [('BART_TEST.', '3fb8369208'), ('BART_TEST ', '077979f009'), ]

    # Calculate the SHA3-512 hash
    hash_factory = HashFactory()
    hash_method = hash_factory.create_hash_method("sha3", 512)

    # Verify the calculated hashes match the expected hashes
    for data, expected_hash in dataset:
        hash_digest = hash_method.calculate_hash(data)[:10]
        assert expected_hash == hash_digest


def test_short_md5():
    dataset = [('BART_TEST ', '1080702796'), ('BART_TEST.', '2994824863')]

    # Calculate the icometrix MD5 hash (MD5 in decimal format)
    hash_factory = HashFactory()
    hash_method = hash_factory.create_hash_method("short_md5")

    # Verify the calculated hash matches the expected hash
    for data, expected_hash in dataset:
        hash_digest = hash_method.calculate_hash(data)
        assert expected_hash == hash_digest, 'hash verification failed'



import logging
import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from time import sleep

from requests.adapters import DEFAULT_POOLSIZE

from icometrix_sdk.exceptions import IcometrixException, IcometrixDataImportException
from icometrix_sdk.logger import logger_name
from icometrix_sdk.models.base import PaginatedResponse
from icometrix_sdk.models.upload_entity import StartUploadDto, UploadEntity, UploadEntityFiles, StudyUploadEntity
from icometrix_sdk.utils.requests_api_client import ApiClient

logger = logging.getLogger(logger_name)


class Uploads:
    def __init__(self, api: ApiClient, polling_interval=2):
        self._api = api
        self.polling_interval = polling_interval

    def get_all(self, project_id: str, **kwargs) -> PaginatedResponse[UploadEntity]:
        """
        List all uploads within a project

        :param: The ID of the project
        :return: A Paginated response containing upload entries
        """
        page = self._api.get(f"/uploads-service/api/v1/projects/{project_id}/dicom-uploads", **kwargs)
        return PaginatedResponse[UploadEntity](**page)

    def get_one(self, upload_uri: str) -> UploadEntity:
        """
        Get a single upload entry based on the upload uri

        :param upload_uri: the uri of the upload
        :return: A single patient or 404
        """
        resp = self._api.get(upload_uri)
        return UploadEntity(**resp)

    def get_studies_for_upload(self, upload_folder_uri: str, **kwargs) -> PaginatedResponse[StudyUploadEntity]:
        """
        Get studies created by this upload (first wait till import is finished)

        :param upload_folder_uri: the folder uri of the upload (UploadEntry.folder_uri)
        :return:
        """
        page = self._api.get(f"{upload_folder_uri}/study-uploads", **kwargs)
        return PaginatedResponse[StudyUploadEntity](**page)

    def get_uploaded_files(self, upload_folder_uri: str, **kwargs) -> UploadEntityFiles:
        """
        List all files that have been uploaded to a UploadEntry
        Once the upload has been completed, and imported the files will be removed

        For name conflict avoidance; the original file name will be prepended with a UUID and sanitized:
        Only ASCII alphanumerals, ``'.'``, ``'-'`` and ``'_'`` are allowed for
        maximum portability and safety wrt using this name as a filename on a
        regular file system. All other characters will be replaced with an
        underscore (``'_'``).

        The `filename` is normalized to the Unicode ``NKFD`` form prior to
        ASCII conversion in order to extract more alphanumerals where a
        decomposition is available. For instance:

        'Bold Digit 𝟏'  => 'Bold_Digit_1'
        'Ångström unit physics.pdf' => 'A_ngstro_m_unit_physics.pdf'

        :param upload_folder_uri: the folder uri of the upload (UploadEntry.folder_uri)
        :return: An object containing all file names (If the upload has been imported, the file list will be empty)
        """
        resp = self._api.get(f"{upload_folder_uri}/files", **kwargs)
        return UploadEntityFiles(**resp)

    def upload_dicom_dir(self, project_id: str, dicom_dir_path: str, options: StartUploadDto,
                         complete_on_error=False) -> UploadEntity:
        """
        A higher level function to upload all DICOMs of a directory AND all of its subdirectories

        :param project_id:
            The ID of the project you want to upload to
        :param dicom_dir_path:
            The path to the directory
        :param options:
            Extra upload options
        :param complete_on_error:
            Setting this boolean to true will still complete the upload, even if a file failed to upload
        :return:
        """
        # Create upload entry
        upload = self.start_upload(project_id, options)
        # Upload all files in directory
        self.upload_all_files_in_dir(upload.uri, dicom_dir_path, complete_on_error)
        # Once all files have been uploaded, signal that they are all there and start the import/processing
        return self.complete_upload(upload.uri)

    def upload_all_files_in_dir(self, upload_uri: str, dicom_dir_path: str, complete_on_error=False) -> int:
        """
        A higher level function to upload all DICOMs of a directory AND all of its subdirectories

        :param upload_uri:
            the URI of the upload entry
        :param dicom_dir_path:
            The path to the directory
        :param complete_on_error:
            Setting this boolean to true will still complete the upload, even if a file failed to upload
        :return:
        """
        futures = []
        with ThreadPoolExecutor(DEFAULT_POOLSIZE) as executor:
            # Loop over all files in a DIR and upload them to the before created upload entry
            for path, subdirs, files in os.walk(dicom_dir_path):
                for name in files:
                    if name.startswith("."):
                        continue

                    file_path = os.path.join(path, name)
                    logger.info(f"Uploading {file_path}")
                    future = executor.submit(self.upload_dicom_path, upload_uri, file_path)
                    futures.append(future)

            # Wait for all threads to complete and handle exceptions
            for completed_future in as_completed(futures):
                try:
                    completed_future.result()  # Check for exceptions here
                except IcometrixException as e:
                    # Log or handle the exception appropriately
                    logging.error(f"Exception in one of the threads: {e}")
                    if not complete_on_error:
                        raise e
        return len(futures)

    def start_upload(self, project_id: str, start_upload: StartUploadDto, **kwargs) -> UploadEntity:
        """
        A function to create an upload entry, that can be used to upload a block of data

        :param project_id: The ID of the project you want to upload to
        :param start_upload: Extra upload options
        :return:
        """
        body = start_upload.model_dump(by_alias=True)
        resp = self._api.post(f"/uploads-service/api/v1/projects/{project_id}/multi-upload", data=body, **kwargs)
        return UploadEntity(**resp)

    def upload_dicom_path(self, upload_uri: str, file_path: str):
        """
        Upload a file to an upload entry

        :param upload_uri: Uri to upload to (UploadEntity.uri)
        :param file_path: The path to the file
        :return:
        """
        with open(file_path, "rb") as fp:
            self.upload_dicom(upload_uri, fields={"file": (file_path, fp.read(), "application/octet-stream")})

    def upload_dicom(self, upload_uri: str, fields):
        """
        Upload files to an upload entry, these can be DICOMs or zips

        :param upload_uri: The uri to upload to
        :param fields: Dictionary of fields or list of (key, :class:`~urllib3.fields.RequestField`).
        :return:
        """
        self._api.put_file(upload_uri, fields=fields)

    def complete_upload(self, upload_uri: str) -> UploadEntity:
        """
        A function to complete an upload entry, and start the importing the files

        :param upload_uri: The uri to upload to
        :return:
        """
        resp = self._api.post(upload_uri, data={})
        return UploadEntity(**resp)

    def wait_for_data_import(self, upload_uri: str) -> UploadEntity:
        """
        After data has been upload (and completed), we need to wait for the data to be imported

        :param upload_uri: The uri to upload to
        :return: UploadEntity
        """

        upload = self.get_one(upload_uri)
        while upload.status != "import_success":
            upload = self.get_one(upload.folder_uri)
            logger.info(upload)
            if upload.status == "import_failed":
                raise IcometrixDataImportException(f"Import failed: {upload}")
            sleep(self.polling_interval)
        return upload



from typing import Optional

from icometrix_sdk.models.base import BackendEntity, utc_datetime


class ProjectEntity(BackendEntity):
    name: str
    state: Optional[str] = None
    abbrev: str
    country: str
    description: Optional[str] = None
    display_name: Optional[str] = None
    project_type: Optional[str] = None
    processing_type: Optional[str] = None
    imported_timestamp: Optional[utc_datetime] = None
